from fam.blud import GenericObject, StringField, ReferenceFrom, ReferenceTo, BoolField, NumberField, DictField, ObjectField, ListField
from fam.string_formats import DateTimeField
from fam.string_formats import EmailField

NAMESPACE = "glowinthedark.co.uk/test"


class Dog(GenericObject):

    additional_properties = True
    sync_gateway_write = True
    fields = {
        "name": StringField(),
        "owner_id": ReferenceTo(NAMESPACE, "person", cascade_delete=True),
        "kennel_club_membership": StringField(unique=True),
        "channels": ListField(default=["callbacks"])
        }

    def talk(self):
        return "woof"

    def pre_save_new_cb(self, db):
        pass

    def post_save_new_cb(self, db):
        pass

    def pre_save_update_cb(self, db, old_properties):
        pass

    def post_save_update_cb(self, db):
        pass

    def pre_delete_cb(self, db):
        pass

    def post_delete_cb(self, db):
        pass

    def changes_new_cb(self, db):
        if self.owner:
            self.owner.add_callback(db, "changes_new_cb")

    def changes_update_cb(self, db):
        if self.owner:
            self.owner.add_callback(db, "changes_update_cb")




class JackRussell(Dog):
    fields = {
        "age": NumberField()
        }

    def talk(self):
        return "Yap"


class Cat(GenericObject):
    fields = {
        "name": StringField(),
        "colour": StringField(immutable=True),
        "tail": BoolField(immutable=True, default=True),
        "legs": NumberField(required=True),
        "owner_id": ReferenceTo(NAMESPACE, "person", required=True),
        "email": EmailField(default="cat@home.com")
        }


    @classmethod
    def all_with_n_legs(cls, db, legs):
        return db.view("animal_views/cat_legs", key=legs)


class Person(GenericObject):
    fields = {
        "name": StringField(),
        "cats": ReferenceFrom(NAMESPACE, "cat", "owner_id", cascade_delete=True),
        "dogs": ReferenceFrom(NAMESPACE, "dog", "owner_id"),
        "animals": ReferenceFrom(NAMESPACE, ["dog", "cat"], "owner_id"),
        "callbacks": ListField()
        }

    def add_callback(self, db, name):

        if self.callbacks is None:
            self.callbacks = []

        self.callbacks.append(name)
        self.save(db)


class Monarch(Person):
    fields = {
        "country": StringField(),
        }


class Monkey(GenericObject):
    use_rev = False
    fields = {
        "name": StringField(),
        "colour": StringField(immutable=True),
        }


class Weapons(object):

    def __init__(self, wings, fire, claws):

        self.fire = fire
        self.claws = claws
        self.wings = wings

    def to_json(self):

        return {
            "fire": self.fire,
            "claws": self.claws,
            "wings": self.wings,
        }

    @classmethod
    def from_json(cls, as_json):
        return cls(as_json["wings"], as_json["fire"], as_json["claws"])



class Monster(GenericObject):

    fields = {
        "name": StringField(),
        "weapons": ObjectField(cls=Weapons),

        }

class Event(GenericObject):

    fields = {
        "name": StringField(),
        "created": DateTimeField()
        }


from fam.blud import ReferenceFrom, GenericObject
import json

class FamDbAuthException(Exception):
    pass

class BaseDatabase(object):

    FOREIGN_KEY_MAP_STRING = '''function(doc) {
    var resources = %s;
    if (resources.indexOf(doc.type) != -1 && doc.namespace == \"%s\"){
        emit(doc.%s, doc);
    }
}'''


###################################

    #double dispatch accessors that return objects

    def put(self, thing):
        return thing.save(self)

    def delete(self, thing):
        return thing.delete(self)

    def get(self, key):
        return GenericObject.get(self, key)

    def delete_key(self, key):
        return GenericObject.delete_key(self, key)

    def query_view(self, view_name, **kwargs):
        return GenericObject.view(self, view_name, **kwargs)

    def changes(self, since=None, channels=None, limit=None, feed=None, timeout=None, filter=None):
        return GenericObject.changes(self, since=since, channels=channels, limit=limit, feed=feed, timeout=timeout, filter=filter)

#################################

    def class_for_type_name(self, type_name, namespace_name):
        return self.mapper.get_class(type_name, namespace_name)

    # def _get_design(self, namespace, namespace_name):
    #
    #     views = {}
    #
    #
    #     for type_name, cls in namespace.iteritems():
    #         for field_name, field in cls.cls_fields.iteritems():
    #             if isinstance(field, ReferenceFrom):
    #                 view_key = "%s_%s" % (type_name, field_name)
    #                 # if view_key in ["person_dogs", "person_animals"]:
    #                 views[view_key] = {"map" : self._get_fk_map(field.refcls, field.refns, field.fkey)}
    #
    #             if field.unique:
    #                 view_key = "%s_%s" % (type_name, field_name)
    #                 # if view_key in ["person_dogs", "person_animals"]:
    #                 views[view_key] = {"map": self._get_fk_map(type_name, namespace_name, field_name)}
    #
    #
    #
    #     design = {
    #        "views": views
    #     }
    #
    #     return design
    #
    #
    # def _get_fk_map(self, class_name, namespace, ref_to_field_name):
    #
    #     if isinstance(class_name, list):
    #         class_names = class_name
    #     else:
    #         class_names = [class_name]
    #
    #     all_sub_class_names = set()
    #     for name in class_names:
    #         sub_classes = set(self.mapper.get_sub_class_names(namespace, name))
    #         all_sub_class_names = all_sub_class_names.union(sub_classes)
    #
    #     arrayStr = '["%s"]' % '", "'.join(all_sub_class_names)
    #     return self.FOREIGN_KEY_MAP_STRING % (arrayStr, namespace, ref_to_field_name)



import unittest


from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monarch
from fam.mapper import ClassMapper

class MapperTests(unittest.TestCase):


    def setUp(self):
        self.mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monarch])


    def tearDown(self):
        pass


    def test_sub_class_refs(self):

        self.assertEqual(set(Monarch.fields.keys()), set(["name", "country", "cats", "dogs", "animals", "callbacks"]))
        self.assertEqual(Monarch.cls_fields.keys(), ["country"])





import unittest
import datetime
import pytz

from fam.tests.models.test01 import Event
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper

COUCHDB_URL = "http://localhost:5984"
COUCHDB_NAME = "test"

class MapperTests(unittest.TestCase):


    def setUp(self):
        mapper = ClassMapper([Event])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass


    def test_date_time(self):

        utc = pytz.utc
        birthday = datetime.datetime(year=2015, month=12, day=5, hour=10, minute=10, second=13, microsecond=876532, tzinfo=utc)
        event = Event(created=birthday)
        self.db.put(event)
        timestamp = event._properties["created"]
        self.assertEqual(timestamp, "2015-12-05T10:10:13.876532Z")
        self.assertEqual(event.created, birthday)

    def test_date_time_without_tz(self):

        utc = pytz.utc
        birthday = datetime.datetime(year=2015, month=12, day=5, hour=10, minute=10, second=13, microsecond=876532)
        event = Event(created=birthday)
        self.db.put(event)
        timestamp = event._properties["created"]
        self.assertEqual(timestamp, "2015-12-05T10:10:13.876532Z")
        self.assertEqual(event.created, birthday.replace(tzinfo=utc))




from .couchdb import CouchDBWrapper
from .sync_gateway import SyncGatewayWrapper
try:
    from .couchbase_server import CouchbaseWrapper
except Exception as e:
    print "failed to import couchbase wrapper"

from fam.utils.backoff import http_backoff

@http_backoff
def get_db(db_type,
           mapper,
           host,
           port=None,
           db_name="sync_gateway",
           https=False,
           username=None,
           password=None,
           backoff=False,
           **kwargs):

    if db_type == "sync_gateway":
        url = _get_url(host, 4984 if port is None else port, https, username, password)
        ## kwargs may inc auth_url
        return SyncGatewayWrapper(mapper, url, db_name, username=username, password=password, **kwargs)
    elif db_type == "couchdb":
        url = _get_url(host, 5984 if port is None else port, https, username, password)
        ## kwargs may inc reset, remote_url, continuous
        return CouchDBWrapper(mapper, url, db_name, **kwargs)
    elif db_type == "couchbase":
        ## kwargs may inc read_only
        return CouchbaseWrapper(mapper, host, db_name, **kwargs)
    else:
        raise NotImplementedError("Can't make a database of type %s" % db_type)


def _get_url(host, port, https, username, password):
    if username is None:
        return "%s://%s:%s" % ("https" if https else "http", host, port)
    else:
        return "%s://%s:%s@%s:%s" % ("https" if https else "http", username, password, host, port)


import sys
import os
from fam.database import CouchDBWrapper, get_db
from fam.mapper import ClassMapper
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster
from fam.tests.common import common_test_classes

current_module = sys.modules[__name__]

from fam.tests.test_couchdb.config import *

TEST_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATA_PATH = os.path.join(TEST_DIR, "common", "data")

def iterCouchDBTests():

    for test_class in common_test_classes:

        name = "{}CouchDB".format(test_class.__name__)

        def setUp(self):
            filepath = os.path.join(DATA_PATH, "animal_views.js")
            mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster], designs=[filepath])

            # self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
            self.db = get_db("couchdb", mapper, "localhost", db_name="test", reset=True)
            self.db.update_designs()
            super(self.__class__, self).setUp()

        def tearDown(self):
            pass

        methods = {
            "setUp": setUp,
            "tearDown":tearDown
        }

        setattr(current_module, name, type(name, (test_class,), methods))

iterCouchDBTests()



import time
import subprocess
import unittest
import requests
import json

DB_HOST = "localhost"

SYNC_GATEWAY_PATH = "/usr/local/bin/sync_gateway"
SYNC_GATEWAY_VIEW_URL = "%s/_design/%s/_view/%s?stale=false&key=\"%s\""

DESIGN_ID = "test"
DOG_ID = "dog"
OWNER_ID = "owner"
DB_NAME = "sync_gateway"

ONE_DESIGN = {
    "_id": DESIGN_ID,
    "views": {
        "person_dogs": {
            "map": "function(doc, meta) {\n    var resources = [\"dog\"];\n    if (resources.indexOf(doc.type) != -1){\n        doc._rev = meta.rev;\n        emit(doc.owner_id, doc);\n    }\n}"
        }
    }
}


VIEW_URL = SYNC_GATEWAY_VIEW_URL
DB_PORT = "4985"
DB_URL = "http://%s:%s/%s" % (DB_HOST, DB_PORT, DB_NAME)

DOG = {
    "type": "dog",
    "owner_id": OWNER_ID,
    "name": "rufus"
}

class ViewTests(unittest.TestCase):

    def setUp(self):
        cmd = "{} -log=* -url walrus:".format(SYNC_GATEWAY_PATH)

        time.sleep(0.25)
        self.gateway = subprocess.Popen(cmd, shell=True)
        time.sleep(0.25)


    def tearDown(self):
        # stop the gateway
        self.gateway.kill()


    def put_in_db(self, key, value):
        url = "%s/%s" % (DB_URL, key)
        rsp = requests.put(url,
                           data=json.dumps(value, indent=4, sort_keys=True),
                           headers={"Content-Type": "application/json", "Accept": "application/json"})
        self.assertEqual(rsp.status_code, 201)

    def query_view(self, design_id, view_name, key):
        url = VIEW_URL % (DB_URL, design_id, view_name, key)
        rsp = requests.get(url)
        # print rsp.text
        self.assertEqual(rsp.status_code, 200)
        results = rsp.json()
        return results["rows"]


    def test_view_without_previous_request(self):
        # put the design in the database
        self.put_in_db("_design/%s" % DESIGN_ID, ONE_DESIGN)

        # put the dog doc in the database
        self.put_in_db(DOG_ID, DOG)

        # query the view
        rows = self.query_view(DESIGN_ID, "person_dogs", OWNER_ID)

        # should find 1 row
        self.assertEqual(len(rows), 1)


    def test_view_with_previous_request(self):
        # put the design in the database
        self.put_in_db("_design/%s" % DESIGN_ID, ONE_DESIGN)

        # this previous request causes the next to fail!
        rows = self.query_view(DESIGN_ID, "person_dogs", OWNER_ID)

        # should find 1 row
        self.assertEqual(len(rows), 0)

        # put the dog doc in the database
        self.put_in_db(DOG_ID, DOG)

        # query the view
        rows = self.query_view(DESIGN_ID, "person_dogs", OWNER_ID)

        # should find 1 row
        self.assertEqual(len(rows), 1)






import copy

FIELD_TYPE_LOOKUP = {
    "ObjectField": {},
    "BoolField": {
        "type": "boolean",
    },
    "NumberField": {
        "type": "number",
    },
    "StringField": {
        "type": "string",
    },
    "ListField": {
        "type": "array",
    },
    "DictField": {
        "type": "object",
    },
    "ReferenceTo": {
        "type": "string"
    },
    "EmailField": {
        "type": "string"
    },
    "DateTimeField": {
        "type": "string"
    }
}


def createJsonSchema(fam_class):

    class_name = fam_class.__name__.lower()
    namespace = fam_class.namespace.lower()

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "title": "A Fam object model for class %s:%s" % (fam_class.namespace, fam_class.__name__),
        "type": "object",
        "properties": {
            "type":{
                "type": "string",
                "pattern": class_name
            },
            "namespace": {
                "type": "string",
                "pattern": namespace
            },
            "schema": {
                "type": "string"
            },
            "_deleted": {
                "type": "boolean"
            },
            "_id": {
                "type": "string"
            },
            "_rev": {
                "type": "string"
            }
        },
        "additionalProperties": fam_class.additional_properties,
    }

    required_fields = []

    for name, field in fam_class.fields.iteritems():
        field_class_name = field.__class__.__name__
        if field_class_name != "ReferenceFrom":
            field_dict = copy.deepcopy(FIELD_TYPE_LOOKUP[field_class_name])
            schema["properties"][name] = field_dict
            if hasattr(field, "pattern"):
                field_dict["pattern"] = field.pattern
            if field.required:
                required_fields.append(name)

    if len(required_fields) > 0:
        schema["required"] = required_fields


    return schema


import requests
import time

def http_backoff(func):

    def func_wrapper(*args, **kwargs):

        if kwargs.get("backoff"):
            connected = False
            counter = 0

            # retry with a backoff
            while not connected:
                try:
                    return func(*args, **kwargs)
                except requests.exceptions.ConnectionError as e:
                    counter += 1
                    if counter < 8:
                        nap = 2 ** counter
                        msg = """Failed to connect!.
                         Has failed {} times.
                         Will try again after {} seconds backoff
                         Origional Error: {}""".format(counter, nap, e)
                        print msg
                        time.sleep(nap)
                    else:
                        raise e
        else:
            # print args
            return func(*args, **kwargs)

    return func_wrapper



from copy import deepcopy
import unittest
from mock import patch

from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE

class CallbackBaseTests:

    class CallbackTests(unittest.TestCase):

        db = None

        def test_new_callbacks(self):

            with patch.object(Dog, 'pre_save_new_cb', return_value=None) as mock_pre_new:
                dog = Dog(name="woofer")
                self.db.put(dog)
                mock_pre_new.assert_called_once_with(self.db)

            with patch.object(Dog, 'post_save_new_cb', return_value=None) as mock_post_new:
                dog = Dog(name="woofer")
                self.db.put(dog)
                mock_post_new.assert_called_once_with(self.db)


        def test_update_callbacks(self):

            with patch.object(Dog, 'pre_save_update_cb', return_value=None) as mock_pre_update:
                dog = Dog(name="woofer")
                self.db.put(dog)
                old_properties = deepcopy(dog._properties)
                mock_pre_update.assert_not_called()
                dog.name = "fly"
                self.db.put(dog)
                mock_pre_update.assert_called_once_with(self.db, old_properties)

            with patch.object(Dog, 'post_save_update_cb', return_value=None) as mock_post_update:
                dog = Dog(name="woofer")
                self.db.put(dog)
                mock_post_update.assert_not_called()
                dog.name = "fly"
                self.db.put(dog)
                mock_post_update.assert_called_once_with(self.db)


        def test_delete_callbacks(self):

            with patch.object(Dog, 'pre_delete_cb', return_value=None) as mock_pre_delete:
                dog = Dog(name="woofer")
                self.db.put(dog)
                mock_pre_delete.assert_not_called()
                dog.name = "fly"
                self.db.delete(dog)
                mock_pre_delete.assert_called_once_with(self.db)


            with patch.object(Dog, 'post_delete_cb', return_value=None) as mock_post_delete:
                dog = Dog(name="woofer")
                self.db.put(dog)
                mock_post_delete.assert_not_called()
                dog.name = "fly"
                self.db.delete(dog)
                mock_post_delete.assert_called_once_with(self.db)



import unittest
import time
from fam.exceptions import *
from fam.tests.models.test01 import GenericObject, Dog, Cat, Person, JackRussell, Monkey, Monarch, NAMESPACE

"""
Wrapping the test case in BaseTestCases allows these base test to hide from the test runner and
only be called in the sub classes with specific databases

"""

class BasicBaseTestCases:

    class BasicTests(unittest.TestCase):

        db = None

       # Test the app
        def test_app(self):
            self.assertNotEqual(self.db, None)


        def test_make_an_object(self):
            dog = Dog(name="woofer")
            self.assertEqual(dog.namespace, NAMESPACE)
            self.assertEqual(dog.type, "dog")
            self.assertEqual(dog.name, "woofer")
            self.assertEqual(dog.__class__, Dog)


        def test_make_an_object2(self):
            dog = Dog(name="woofer")
            self.assertEqual(dog.talk(), "woof")


        def test_make_a_sub_object(self):
            jack = JackRussell()
            self.assertEqual(jack.talk(), "Yap")
            jack.name = "jack"
            jack.age = 12
            self.db.put(jack)


        def test_make_an_object_saved(self):
            dog = Dog(name="woofer")
            dog.save(self.db)
            self.assertEqual(dog.namespace, NAMESPACE)
            self.assertEqual(dog.type, "dog")
            self.assertEqual(dog.name, "woofer")
            self.assertEqual(dog.__class__, Dog)
            self.assertNotEqual(dog.key, None)
            # self.assertTrue(dog.schema.startswith("http://glowinthedark.co.uk/test/dog"))


        def test_make_an_object_double_dispatch_saved(self):
            dog = Dog(name="woofer")
            self.db.put(dog)
            self.assertEqual(dog.namespace, NAMESPACE)
            self.assertEqual(dog.type, "dog")
            self.assertEqual(dog.name, "woofer")
            self.assertEqual(dog.__class__, Dog)
            self.assertNotEqual(dog.key, None)


        def test_make_an_object_with_additional_properties(self):
            dog = Dog(name="woofer", collar="leather")
            dog.home = "hackney"
            self.db.put(dog)
            self.assertEqual(dog.home, "hackney")

        def test_fail_with_additional_properties(self):

            def wrong_monkey():
                return Monkey(name="bonzo", collar="leather")

            def collar_monkey(monkey):
                monkey.collar = "leather"

            self.assertRaises(FamValidationError, wrong_monkey)
            monkey = Monkey(name="bonzo")
            self.assertRaises(FamValidationError, collar_monkey, monkey)


        def test_make_an_object_saved_cas(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            self.assertEqual(cat.namespace, NAMESPACE)
            self.assertEqual(cat.type, "cat")
            self.assertNotEqual(cat.rev, None)


        def test_get_cat(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            another = Cat.get(self.db, cat.key)
            self.assertEqual(cat, another)


        def test_ref_to(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            self.assertEqual(cat.owner, paul)
            self.assertEqual(cat.owner.name, "paul")


        def test_ref_to_with_object(self):
            paul = Person(name="paul")
            self.db.put(paul)
            cat = Cat(name="whiskers", owner=paul, legs=2)
            cat.save(self.db)
            self.assertEqual(cat.owner, paul)
            self.assertEqual(cat.owner.name, "paul")
            catback = self.db.get(cat.key)
            self.assertEqual(cat, catback)


        def test_ref_from(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
            cat2.save(self.db)
            self.assertEqual(len(paul.cats), 2)
            self.assertTrue(paul.cats[0] == cat or paul.cats[1] == cat)


        def test_ref_from_multiple_index(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            dog = Dog(name="fly", owner=paul)
            self.db.put(dog)
            self.assertEqual(len(paul.cats), 1)
            self.assertEqual(len(paul.dogs), 1)
            self.assertEqual(len(paul.animals), 2)


        def test_refs_with_inheritance(self):
            paul = Person(name="paul")
            paul.save(self.db)
            jack = JackRussell()
            jack.owner_id = paul.key
            jack.name = "jack"
            jack.save(self.db)
            self.assertEqual(paul.dogs[0], jack)


        def test_refs_with_other_inheritance(self):
            paul = Monarch(name="paul")
            self.db.put(paul)
            jack = Dog()
            jack.owner = paul
            jack.name = "jack"
            self.db.put(jack)
            self.assertEqual(paul.dogs[0], jack)


        def test_delete_cat_dd(self):
            paul = Person(name="paul")
            self.db.put(paul)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            self.db.put(cat)
            key = cat.key
            another = self.db.get(cat.key)
            self.assertEqual(cat, another)
            self.db.delete(cat)
            revivedcat = self.db.get(key)
            self.assertEqual(revivedcat, None)


        def test_delete_dog_to_refs(self):
            paul = Person(name="paul")
            paul.save(self.db)
            dog = Dog(name="rufus", owner_id=paul.key)
            dog.save(self.db)
            key = dog.key
            dog2 = Dog(name="fly", owner_id=paul.key)
            dog2.save(self.db)
            self.assertNotEqual(dog2.owner, None)
            key = paul.key
            dog.delete(self.db)
            revivedpaul = self.db.get(key)
            self.assertEqual(revivedpaul, None)
            refresheddog2 = Dog.get(self.db, dog2.key)
            self.assertEqual(refresheddog2.owner, None)


        def test_delete_cat_refs(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            key = cat.key
            cat2 = Cat(name="puss", owner_id=paul.key, legs=2)
            cat2.save(self.db)
            revivedcat1 = self.db.get(key)
            self.assertNotEqual(revivedcat1, None)
            paul.delete(self.db)
            revivedcat2 = self.db.get(key)
            self.assertEqual(revivedcat2, None)


        def test_delete_dog_refs(self):
            paul = Person(name="paul")
            paul.save(self.db)
            dog = Dog(name="rufus", owner_id=paul.key)
            dog.save(self.db)
            key = dog.key
            dog2 = Dog(name="fly", owner_id=paul.key)
            dog2.save(self.db)
            reviveddog1 = self.db.get(key)
            self.assertNotEqual(reviveddog1, None)
            paul.delete(self.db)
            reviveddog2 = self.db.get(key)
            self.assertNotEqual(reviveddog2, None)


        def test_update_cat(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            cat.name = "blackie"
            cat.save(self.db)
            self.assertEqual(cat.name, "blackie")
            self.assertEqual(cat._properties["name"], "blackie")
            self.assertFalse("name" in cat.__dict__.keys())


        def test_update_cat_fails(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            cat.colour = "black"
            cat.save(self.db)
            def change_colour():
                cat.colour = "white"
            self.assertRaises(FamImmutableError, change_colour)


        def setcatfood(self):
            self.cat.food = "biscuits"


        def test_update_catfood(self):
            paul = Person(name="paul")
            paul.save(self.db)
            cat = Cat(name="whiskers", owner_id=paul.key, legs=2)
            cat.save(self.db)
            self.assertRaises(Exception, self.setcatfood)


        def test_update_dogfood(self):
            dog = Dog(name="fly")
            dog.food = "biscuits"
            dog.save(self.db)


        def test_update_fails_without_rev(self):
            dog = Dog(name="fly")
            dog.save(self.db)
            dog.rev = None
            dog.name = "jess"
            self.assertRaises(FamResourceConflict, self.db.put, dog)


        def test_update_works_without_rev(self):
            monkey = Monkey(name="fly")
            monkey.save(self.db)
            monkey.rev = None
            monkey.name = "jess"
            self.db.put(monkey)

        def test_uniqueness(self):

            paul = Person(name="paul")
            self.db.put(paul)
            dog1 = Dog(name="rufus", owner_id=paul.key, kennel_club_membership="123456")

            dog1.save(self.db)
            # time.sleep(1)


            # print dog1.as_json()
            # dog2 = Dog(name="fly", owner_id=paul.key, kennel_club_membership="123456")
            # print dog2.as_json()
            # # self.db.put(dog2)
            # self.assertRaises(FamUniqueError, self.db.put, dog2)
            # # print "*********** end ***************"

    class RefNameTests(unittest.TestCase):

        # this test that refs whos name doesnt end with _id fail
        def test_misnamed_ref_to_fails(self):

            def duff_import():
                from fam.tests.models import test02
                print test02

            self.assertRaises(FamError, duff_import)


import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper

from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class IndexBaseTests:

    class IndexTests(unittest.TestCase):

        db = None

        # def setUp(self):
        #     filepath = os.path.join(DATA_PATH, "animal_views.js")
        #     mapper = ClassMapper([Dog, Cat, Person, JackRussell], design_js_paths=[filepath])
        #     self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        #     self.db.update_designs()
        #
        # def tearDown(self):
        #     pass

        def test_create_index(self):
            filepath = os.path.join(DATA_PATH, "animal_views.js")

            as_dict = self.db.mapper._js_design_as_doc(filepath)

            expected = {
                "_id": "_design/animal_views",
                "views": {
                    "cat_legs": {
                        "map": "function(doc) {\n  if (doc.type == \"cat\") {\n    emit(doc.legs, doc);\n  }\n}"
                    }
                }
            }

            self.assertEqual(as_dict, expected)


        def test_query_view(self):

            paul = Person(name="Paul")
            self.db.put(paul)
            cat1 = Cat(owner=paul, legs=4)
            self.db.put(cat1)
            cat2 = Cat(owner=paul, legs=3)
            self.db.put(cat2)
            three_legged_cats = Cat.all_with_n_legs(self.db, 3)
            self.assertEqual(len(three_legged_cats), 1)

            self.assertEqual(three_legged_cats[0].key, cat2.key)


        def test_long_polling(self):
            paul = Person(name="Paul")
            self.db.put(paul)
            cat1 = Cat(owner=paul, legs=4)
            self.db.put(cat1)
            cat2 = Cat(owner=paul, legs=3)
            self.db.put(cat2)
            three_legged_cats = self.db.view("animal_views/cat_legs", key=3)
            self.assertEqual(len(three_legged_cats), 1)
            self.assertEqual(three_legged_cats[0].key, cat2.key)


from .blud import StringField
import pytz

import datetime

class EmailField(StringField):
    pattern = """^([-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9a-zA-Z]+)*|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*")@(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,6}\.?$"""


class DateTimeField(StringField):
    pattern = """^([0-9]{4})-(0[1-9]|1[0-2])-([0-2][0-9]|3[0-1])T([0-1][0-9]|2[0-3]):([0-5][0-9]):[0-5][0-9](.[0-9]{1,8})?Z$"""

    @staticmethod
    def to_json(dt):
        # force to utc and then to RFC 3339
        utc = pytz.utc
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=utc)
        else:
            dt = dt.astimezone(utc)
        as_iso = dt.isoformat("T") + "Z"
        #remove unnecessary offset if there is one
        return as_iso.replace("+00:00", "")

    @staticmethod
    def from_json(as_json):
        if "." in as_json:
            dt = datetime.datetime.strptime(as_json, '%Y-%m-%dT%H:%M:%S.%fZ')
        else:
            dt = datetime.datetime.strptime(as_json, '%Y-%m-%dT%H:%M:%SZ')
        dt = dt.replace(tzinfo=pytz.utc)
        return dt



import traceback


class FamWriteBuffer(object):

    def __init__(self, db):
        self.store = {}
        self.to_be_saved = set()
        self.db = db
        self.views = db.mapper.buffer_views


    def _get(self, *args, **kwargs):
        traceback.print_stack()
        raise Exception()


    def view(self, name, **kwargs):
        traceback.print_stack()
        raise NotImplementedError("view doesn't work with the DataBaseCache yet")


    def put(self, thing):
        thing._db = self
        self.views.index_obj(thing)

        key = thing.key
        if key in self.store:
            if id(not self.store[key]) == id(thing):
                raise Exception("putting thing with same key but different python id into cache - it's confused")
        else:
            self.store[thing.key] = thing
        self.to_be_saved.add(key)


    def query_view(self, view_name, **kwargs):
        db_objs = self.db.query_view(view_name, **kwargs)
        [self._refresh_cache(o.key, o) for o in db_objs]
        # these will inclde the refreshed ones
        buffer_objs = self.views.query_view(view_name, **kwargs)
        return buffer_objs


    def _refresh_cache(self, key, got):
        if got is not None:
            got._db = self
        current_in_cache = self.store.get(key)
        # If there are changes from the database then update the current one in place
        if current_in_cache is not None and got is not None:
            if current_in_cache.rev != got.rev:
                current_in_cache.rev = got.rev
                current_in_cache._properties = got._properties
                self.views.index_obj(current_in_cache)
            return current_in_cache
        elif current_in_cache is not None:
            return current_in_cache
        elif got is not None:
            self.store[got.key] = got
            self.views.index_obj(got)
            return got
        else:
            return None


    def delete(self, thing):
        if thing.key in self.store:
            del self.store[thing.key]
        return self.db.delete(thing)


    def get(self, key):
        got = self.db.get(key)
        return self._refresh_cache(key, got)


    def delete_key(self, key):
        if key in self.store:
            del self.store[key]
        return self.db.delete_key(key)


    def flush(self):
        # only save the things that have been "put"
        for key in list(self.to_be_saved):
            thing = self.store[key]
            thing._db = self.db
            self.db.put(thing)

        self.views.clear_indexes()



from __future__ import absolute_import
import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper
from fam.buffer import buffered_db
from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class CacheTests(unittest.TestCase):

    def setUp(self):
        mapper = ClassMapper([Dog, Cat, Person])
        self.db = CouchDBWrapper(mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass


    def test_cache_saves(self):

        with buffered_db(self.db) as dbc:
            dog = Dog(name="woofer")
            dbc.put(dog)

        got = self.db.get(dog.key)

        self.assertTrue(got != None)


    def test_cache_doesnt_save(self):
        # doesnt save until were done

        with buffered_db(self.db) as dbc:
            dog = Dog(name="woofer")
            dbc.put(dog)
            got = self.db.get(dog.key)
            self.assertTrue(got is None)

        got = self.db.get(dog.key)
        self.assertTrue(got is not None)

    def test_cache_gets(self):
        # doesnt save until were done

        with buffered_db(self.db) as dbc:
            dog = Dog(name="woofer")
            dbc.put(dog)
            fetched = dbc.get(dog.key)
            self.assertTrue(fetched is not None)
            self.assertEqual(id(dog), id(fetched))


    def test_cache_gets_from_db(self):

        dog = Dog(name="woofer")
        self.db.put(dog)

        with buffered_db(self.db) as dbc:
            fetched = dbc.get(dog.key)
            self.assertTrue(fetched is not None)
            self.assertNotEqual(id(dog), id(fetched))
            fetched_again = dbc.get(dog.key)

            self.assertEqual(id(fetched), id(fetched_again))

    def test_cache_gets_change_from_db(self):

        dog = Dog(name="woofer")
        self.db.put(dog)

        with buffered_db(self.db) as dbc:
            fetched = dbc.get(dog.key)
            self.assertTrue(fetched is not None)
            self.assertNotEqual(id(dog), id(fetched))
            fetched_again = dbc.get(dog.key)
            self.assertTrue(fetched_again is not None)

            self.assertEqual(id(fetched), id(fetched_again))
            dog.name = "fly"
            self.db.put(dog)

            fetched_yet_again = dbc.get(dog.key)
            self.assertTrue(fetched_yet_again is not None)

            self.assertEqual(id(fetched), id(fetched_yet_again))
            self.assertEqual(fetched_yet_again.name, 'fly')
            fetched_yet_again.name = "bluebottle"
            db_fetched = self.db.get(dog.key)
            self.assertEqual(db_fetched.name, 'fly')

        db_fetched = self.db.get(dog.key)
        self.assertEqual(db_fetched.name, 'fly')


    def test_saves_putted(self):

        dog = Dog(name="woofer")
        self.db.put(dog)

        with buffered_db(self.db) as dbc:
            fetched = dbc.get(dog.key)
            fetched.name = "bluebottle"
            dbc.put(fetched)

        db_fetched = self.db.get(dog.key)
        self.assertEqual(db_fetched.name, 'bluebottle')



    def test_refs_from(self):

        with buffered_db(self.db) as dbc:

            person = Person(name="paul")
            dbc.put(person)

            dog = Dog(name="woofer", owner=person)
            dbc.put(dog)

            self.assertEqual(person.dogs, [dog])





import requests
import json
import time
import subprocess

def make_a_bucket(couchbase_url, user_name, password, bucket_name, force=False, flush=False):

    params = {}
    params["authType"] = "none"
    params["proxyPort"] = "11224"
    params["bucketType"] = "couchbase"
    params["flushEnabled"] = "1"
    params["name"] = bucket_name
    params["ramQuotaMB"] = "128"
    params["replicaNumber"] = "0"

    if flush:
        params["flushEnabled"] = "1"

    headers = {'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8'}

    rsp = requests.post("%s/pools/default/buckets" % (couchbase_url), data=params, auth=(user_name, password))

    if rsp.status_code == 400:
        info = json.loads(rsp.text)
        errors = info["errors"]
        name_error = errors.get("name")
        if name_error == "Bucket with given name already exists":
            if force:
                delete_a_bucket(couchbase_url, user_name, password, bucket_name)
                return make_a_bucket(couchbase_url, user_name, password, bucket_name)
            else:
                raise Exception("failed to make new bucket %s : %s" % (rsp.status_code, rsp.text))

    if rsp.status_code != 202:
        raise Exception("failed to make new bucket %s : %s" % (rsp.status_code, rsp.text))


def delete_a_bucket(couchbase_url, user_name, password, bucket_name):

    time.sleep(1)

    rsp = requests.delete("%s/pools/default/buckets/%s" % (couchbase_url, bucket_name), auth=(user_name, password))

    if rsp.status_code != 200:
        raise Exception("failed to delete bucket %s : %s" % (rsp.status_code, rsp.text))


def number_of_buckets(couchbase_url, user_name, password):

    rsp = requests.get("%s/pools/default/buckets" % (couchbase_url), auth=(user_name, password))

    if rsp.status_code != 200:
        raise Exception("failed get bucket count %s : %s" % (rsp.status_code, rsp.text))

    buckets = rsp.json()

    print buckets
    return len(buckets)


def flush_a_bucket(couchbase_url, user_name, password, bucket_name):

    rsp = requests.post("%s/pools/default/buckets/%s/controller/doFlush" % (couchbase_url, bucket_name), auth=(user_name, password))

    # if rsp.status_code != 200:
    #     raise Exception("failed to flush bucket %s : %s" % (rsp.status_code, rsp.text))


def make_a_gateway(sync_admin_url, db_name, couchbase_url, bucket, sync_function, force=False):

    config = {db_name:{
         "server": couchbase_url,
         "bucket": bucket,
         "sync": sync_function}
        }


    rsp = requests.put("%s/%s/" % (sync_admin_url, db_name), data=json.dumps(config))

    if rsp.status_code == 412:
        info = json.loads(rsp.text)
        error = info.get("error")
        reason = info.get("reason")
        if reason and reason.startswith("Duplicate database name") and force:
            delete_a_gateway(sync_admin_url, db_name)
            return make_a_gateway(sync_admin_url, db_name, couchbase_url, bucket, sync_function)
        else:
            raise Exception("failed to make a new gateway %s : %s" % (rsp.status_code, rsp.text))

    if rsp.status_code != 201:
        raise Exception("failed to make a new gateway %s : %s" % (rsp.status_code, rsp.text))


def delete_a_gateway(sync_admin_url, db_name):

    rsp = requests.delete("%s/%s/" % (sync_admin_url, db_name))

    if rsp.status_code != 200:
        raise Exception("failed to make a new gateway %s : %s" % (rsp.status_code, rsp.text))



def does_person_exist(sync_admin_url, db_name, username):

    rsp = requests.get("%s/%s/_user/%s" % (sync_admin_url, db_name, username))

    if rsp.status_code == 200:
        return True
    elif rsp.status_code == 404:
        return False
    else:
        raise Exception("failed to add person %s : %s" % (rsp.status_code, rsp.text))



def add_person_to_gateway(sync_admin_url, db_name, user_id, username, password, domain_role=None, admin_channels=None):

    if sync_admin_url is None:
        return

    attrs = {
         "password": password,
         "admin_roles": [user_id]
        }

    if domain_role is not None:
        attrs["admin_roles"].append(domain_role)

    if admin_channels is not None:
        attrs["admin_channels"] = admin_channels

    rsp = requests.put("%s/%s/_user/%s" % (sync_admin_url, db_name, username), data=json.dumps(attrs))

    if rsp.status_code >= 300:
        raise Exception("failed to add person %s : %s" % (rsp.status_code, rsp.text))



def add_guest_to_gateway(sync_admin_url, db_name):

    rsp = requests.put("%s/%s/_user/GUEST" % (sync_admin_url, db_name), data='{"disabled":false, "admin_channels":["public"]}')

    if rsp.status_code != 200:
        pass
        #raise Exception("failed to add user %s : %s" % (rsp.status_code, rsp.text))



def make_bucket_and_gateway(couchbase_url,
                            couchbase_user_name,
                            couchbase_password,
                            bucket_name,
                            sync_admin_url,
                            sync_db_name,
                            sync_function,
                            guest=False,
                            force=False):

    make_a_bucket(couchbase_url, couchbase_user_name, couchbase_password, bucket_name, force=force)
    # make_a_gateway(sync_admin_url, sync_db_name, couchbase_url, bucket_name, sync_function, force=force)
    #if guest:
    #    add_guest_to_gateway(sync_admin_url, sync_db_name)


def delete_bucket_and_gateway(couchbase_url,
                            couchbase_user_name,
                            couchbase_password,
                            bucket_name,
                            sync_admin_url,
                            sync_db_name):


    # delete_a_gateway(sync_admin_url, sync_db_name)
    delete_a_bucket(couchbase_url, couchbase_user_name, couchbase_password, bucket_name)




from __future__ import absolute_import
import os
import unittest
from fam.database import CouchDBWrapper
from fam.mapper import ClassMapper
from fam.buffer.buffer_views import FamWriteBufferViews
from fam.tests.test_couchdb.config import *
from fam.tests.models.test01 import Dog, Cat, Person

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
DATA_PATH = os.path.join(THIS_DIR, "data")

class CacheTests(unittest.TestCase):


    def setUp(self):
        self.mapper = ClassMapper([Dog, Cat, Person])
        self.db = CouchDBWrapper(self.mapper, COUCHDB_URL, COUCHDB_NAME, reset=True)
        self.db.update_designs()

    def tearDown(self):
        pass


    def test_make_views(self):

        views = FamWriteBufferViews(self.mapper)
        paul = Person(name="paul")
        dog = Dog(name="woofer", owner=paul)
        views.index_obj(dog)
        self.assertTrue(views.indexes.get("glowinthedark_co_uk_test_person_dogs") != None)
        self.assertTrue(views.indexes["glowinthedark_co_uk_test_person_dogs"][paul.key][dog.key] == dog)


    def test_query_views(self):

        views = FamWriteBufferViews(self.mapper)
        paul = Person(name="paul")
        dog = Dog(name="woofer", owner=paul)
        views.index_obj(dog)
        obj = views.query_view("glowinthedark.co.uk/test/person_dogs", key=paul.key)
        self.assertEqual(obj, [dog])


    def test_views_keys(self):

        views = FamWriteBufferViews(self.mapper)
        paul = Person(name="paul")
        dog = Dog(name="woofer", owner=paul)
        views.index_obj(dog)
        self.assertEqual(views.indexes.keys(), ['glowinthedark_co_uk_test_person_animals', 'glowinthedark_co_uk_test_person_dogs', 'raw_all', 'glowinthedark_co_uk_test_dog_kennel_club_membership'])







from fam.blud import GenericObject, StringField, ReferenceFrom, ReferenceTo, BoolField, NumberField, DictField
from fam.string_formats import EmailField

NAMESPACE = "glowinthedark.co.uk/test/3"


class Dog(GenericObject):
    additional_properties = True
    fields = {
        "name": StringField(default="fly", required=True),
        "owner_id": ReferenceTo(NAMESPACE, "person", cascade_delete=True)
        }

    def talk(self):
        return "woof"





import sys

SYNC_GATEWAY_HOST = "localhost"
SYNC_GATEWAY_ADMIN_HOST = "localhost"
SYNC_GATEWAY_PORT = "4984"
SYNC_GATEWAY_ADMIN_PORT = "4985"
SYNC_GATEWAY_NAME = "sync_gateway"

if sys.platform.startswith('linux'):
    SYNC_GATEWAY_PATH = "/opt/couchbase-sync-gateway/bin/sync_gateway"
else:
    # SYNC_GATEWAY_PATH = "/usr/local/bin/sync_gateway"
    SYNC_GATEWAY_PATH = "/Users/paul/Dropbox/glowinthedark/spate/Flotsam/bin/sync_gateway_versions/1.3/sync_gateway"



import unittest
import requests
from config import *
import time

from fam.utils import couchbase_utils

class CouchbaseTests(object):


    def test_talk_to_server(self):
        rsp = requests.get("%s/pools/nodes" % (COUCHBASE_LOCAL_URL))
        self.assertTrue(rsp.status_code < 300)


    def test_make_and_delete_a_bucket(self):

        bucket_name = "test_bucket"
        origional_count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)

        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count + 1)


        couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count)


    def test_make_flush_a_bucket(self):

        bucket_name = "test_bucket"
        couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)


        origional_count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)

        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count + 1)

        time.sleep(1)

        couchbase_utils.flush_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)


        couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count)


    def test_ensure_empty_bucket(self):

        bucket_name = "test_bucket"
        origional_count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)


        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count + 1)

        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name, force=True)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count + 1)

        couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)
        count = couchbase_utils.number_of_buckets(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD)
        self.assertEquals(count, origional_count)


class SyncGatewayTests(unittest.TestCase):


    def test_make_a_gateway(self):

        bucket_name = "test_bucket"

        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name, force=True)

        try:
            couchbase_utils.make_a_gateway(COUCHDB_SYNC_ADMIN_URL, "test_gateway", COUCHBASE_LOCAL_URL, bucket_name, "function(doc) {channel(doc.channels);}", force=True)
            couchbase_utils.delete_a_gateway(COUCHDB_SYNC_ADMIN_URL, "test_gateway")

        finally:
            couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)



    def test_add_guest(self):

        bucket_name = "test_bucket"

        couchbase_utils.make_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name, force=True)


        try:
            couchbase_utils.make_a_gateway(COUCHDB_SYNC_ADMIN_URL, "test_gateway", COUCHBASE_LOCAL_URL, bucket_name, "function(doc) {channel(doc.channels);}", force=True)
            couchbase_utils.add_guest_to_gateway(COUCHDB_SYNC_ADMIN_URL, "test_gateway")
            couchbase_utils.delete_a_gateway(COUCHDB_SYNC_ADMIN_URL, "test_gateway")

        finally:
            couchbase_utils.delete_a_bucket(COUCHBASE_LOCAL_URL, COUCHBASE_ADMIN, COUCHBASE_ADMIN_PASSWORD, bucket_name)



class BigFatTest(unittest.TestCase):

    def test_making_both_bucket_and_gateway(self):

        bucket_name = "test_bucket"
        gateway_name = "test_gateway"

        try:
            couchbase_utils.make_bucket_and_gateway(COUCHBASE_LOCAL_URL,
                                COUCHBASE_ADMIN,
                                COUCHBASE_ADMIN_PASSWORD,
                                bucket_name,
                                COUCHDB_SYNC_ADMIN_URL,
                                gateway_name,
                                "function(doc) {channel(doc.channels);}",
                                guest=True,
                                force=True)



        finally:
            couchbase_utils.delete_bucket_and_gateway(COUCHBASE_LOCAL_URL,
                                COUCHBASE_ADMIN,
                                COUCHBASE_ADMIN_PASSWORD,
                                bucket_name,
                                COUCHDB_SYNC_ADMIN_URL,
                                gateway_name)


from setuptools import setup, find_packages

setup(name='fam',
    version='1.1.3',
    description="Simple Python ORM for CouchDB, and Sync Gateway",
    url="https://github.com/paulharter/fam",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'License :: OSI Approved :: MIT License'
    ],
    author='Paul Harter',
    author_email='paul@glowinthedark.co.uk',
    license="LICENSE",
    install_requires=['js2py', 'requests', 'simplejson', 'jsonschema', 'mock', 'pytz', 'slimit', 'ply==3.4'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False)





from .requirement import CreateRequirement, DeleteRequirement, UpdateRequirement

ANYONE = None
NO_ONE = []

__all__ = [
    "CreateRequirement",
    "UpdateRequirement",
    "DeleteRequirement",
    "ANYONE",
    "NO_ONE"
]



import simplejson as json
import hashlib
import copy
from fam.fam_json import object_default

from base64 import b64encode


from fam.utils import requests_shim as requests

from .couchdb import CouchDBWrapper, ResultWrapper

class SyncGatewayWrapper(CouchDBWrapper):

    ## the option stale=false forces the view to be indexed on read. Sync_gateway does not index on write!!
    # VIEW_URL = "%s/%s/_design/%s/_view/%s?stale=false&key=\"%s\""

    # this function is different from the base version in that it adds the rev from the meta into the doc
    FOREIGN_KEY_MAP_STRING = '''function(doc, meta) {
    var resources = %s;
    if (resources.indexOf(doc.type) != -1 && doc.namespace == \"%s\"){
        doc._rev = meta.rev;
        emit(doc.%s, doc);
    }
}'''

    database_type = "sync_gateway"

    def __init__(self, mapper, db_url, db_name,
                 auth_url=None,
                 username=None,
                 password=None,
                 validator=None,
                 read_only=False):


        self.mapper = mapper
        self.validator = validator
        self.read_only = read_only

        self.db_name = db_name
        self.db_url = db_url
        self.username = username
        self.password = password
        self.auth_url = auth_url
        self.session = requests.Session()

        url = "%s/%s" % (db_url, db_name)

        print url
        rsp = self.session.get(url)

        self.cookies = {}

        if rsp.status_code == 404:
            raise Exception("Unknown database and you can't create them in the sync gateway")

    def authenticate(self):
        if self.username is None:
            raise Exception("failed to authenticate no username")

        userAndPass = b64encode(b"paul:bumbum").decode("ascii")
        headers = { 'Authorization' : 'Basic %s' %  userAndPass }

        rsp = self.session.get(self.auth_url, headers=headers)
        if rsp.status_code == 200:
            self.cookies = rsp.cookies
        else:
            raise Exception("failed to authenticate")


    def _wrapper_from_view_json(self, as_json):
        return ResultWrapper.from_gateway_view_json(as_json)


    # def changes(self, since=None, channels=None, limit=None):
    #     raise NotImplementedError("Haven't done changes for sync gateway yet")


    def sync_up(self):
        pass

    def sync_down(self):
        pass


    def purge(self, key):

        if self.read_only:
            raise Exception("This db is read only")

        data = {
            key: ["*"]
        }

        rsp = self.session.post("%s/%s/_purge" % (self.db_url, self.db_name), data=json.dumps(data))

        if rsp.status_code == 200 or rsp.status_code == 202:
            return


    def user(self, username):
        url = "%s/%s/_user/%s" % (self.db_url, self.db_name, username)
        rsp = self.session.get(url)
        if rsp.status_code == 200:
            return rsp.json()
        else:
            return None

    def role(self, role_name):
        url = "%s/%s/_role/%s" % (self.db_url, self.db_name, role_name)
        rsp = self.session.get(url)
        if rsp.status_code == 200:
            return rsp.json()
        else:
            return None


    def ensure_role(self, role_name):

        role_info = self.role(role_name)
        if role_info is None:
            data = {
                "name": role_name,
                "db": self.db_name
            }

            url = "%s/%s/_role/%s" % (self.db_url, self.db_name, role_name)

            rsp = self.session.put(url, data=json.dumps(data, indent=4, sort_keys=True, default=object_default),
                                   headers={"Content-Type": "application/json", "Accept": "application/json"})
            if rsp.status_code == 200 or rsp.status_code == 201:
               return True
            else:
                return False
        else:
            return True


    def ensure_user_role(self, username, role):

        user_info = self.user(username)
        roles = user_info["admin_roles"]

        if not role in roles:
            roles.append(role)

            url = "%s/%s/_user/%s" % (self.db_url, self.db_name, username)

            rsp = self.session.put(url, data=json.dumps(user_info, indent=4, sort_keys=True, default=object_default),
                                   headers={"Content-Type": "application/json", "Accept": "application/json"})
            if rsp.status_code == 200 or rsp.status_code == 201:
               return True
            else:
                return False
        else:
            return True


    def view(self, name, **kwargs):
        return super(SyncGatewayWrapper, self).view(name, stale="false", **kwargs)

    # @auth
    def get_design(self, key):

        url = "%s/%s/%s" % (self.db_url, self.db_name, key)
        print url
        rsp = self.session.get(url)

        if rsp.status_code == 200:
            return rsp.json()
        if rsp.status_code == 500:
            return None
        if rsp.status_code == 400:
            return None
        if rsp.status_code == 404:
            return None
        raise Exception("Unknown Error getting cb doc: %s %s" % (rsp.status_code, rsp.text))


    def _new_matches_existing(self, new_doc, existing_doc):

        new_view_names = new_doc["views"].keys()
        existing_view_names = existing_doc["views"].keys()

        # print "new_view_names: ", new_view_names
        # print "existing_view_names: ", existing_view_names

        if set(new_view_names) != set(existing_view_names):
            # print "names dont match"
            return False

        for view_name in new_view_names:
            new_view_function = new_doc["views"][view_name]["map"]
            existing_view_function = existing_doc["views"][view_name]["map"]
            index = existing_view_function.find(new_view_function)
            if index == -1:
                # print "functions dont match"
                return False

        return True


    def ensure_design_doc(self, key, doc):

        if self.read_only:
            raise Exception("This db is read only")

        existing = self.get_design(key)

        if existing is None or not self._new_matches_existing(doc, existing):
            print "************  updating design doc %s ************" % key
            print "new_design: ", doc
            self._set(key, doc)
        else:
            print "************  design doc %s up to date **********" % key


    def _raw_design_doc(self):

        design_doc = {
            "views": {
                "all": {
                    "map": """function(doc, meta) {
                        doc._rev = meta.rev;
                        emit(doc.type, doc);
                    }
                    """
                }
            }
        }

        return design_doc



import sys
import os
import time
import subprocess
from fam.database import SyncGatewayWrapper
from fam.mapper import ClassMapper
from fam.tests.test_sync_gateway.config import *
from fam.tests.models.test01 import Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster
from fam.tests.common import common_test_classes

current_module = sys.modules[__name__]

TEST_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
DATA_PATH = os.path.join(TEST_DIR, "common", "data")

def iterSyncGatewayTests():

    for test_class in common_test_classes:
        name = "{}SyncGateway".format(test_class.__name__)

        def setUp(self):

            cmd = "{} -log=* -url walrus: ".format(SYNC_GATEWAY_PATH)
            print cmd

            time.sleep(0.25)
            self.gateway = subprocess.Popen(cmd, shell=True)
            time.sleep(0.25)
            filepath = os.path.join(DATA_PATH, "animal_views.js")
            mapper = ClassMapper([Dog, Cat, Person, JackRussell, Monkey, Monarch, Monster], designs=[filepath])

            url = "http://%s:%s" % (SYNC_GATEWAY_ADMIN_HOST, SYNC_GATEWAY_ADMIN_PORT)
            self.db = SyncGatewayWrapper(mapper, url, SYNC_GATEWAY_NAME)
            self.db.update_designs()
            super(self.__class__, self).setUp()

        def tearDown(self):
            # stop the gateway
            self.gateway.kill()

        methods = {
            "setUp": setUp,
            "tearDown":tearDown
        }

        setattr(current_module, name, type(name, (test_class,), methods))


## not running the gateway tests for circle as walrus doesnt work properly there
iterSyncGatewayTests()

